#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <pthread.h>
#include <strings.h>
#include <gtk/gtk.h>
#include "epos_protocol.h"

#define LISTEN_INTERVAL 20
static char buf[BUFFER_SIZE];

static char * green_led_xpm[] = {
"16 16 33 1",
"       c None",
".      c #030000",
"+      c #064702",
"@      c #169302",
"#      c #16AF02",
"$      c #18C801",
"%      c #04E348",
"&      c #4BFCA2",
"*      c #C3FCDF",
"=      c #18EB02",
"-      c #04EB7C",
";      c #03FC7E",
">      c #1BFC8A",
",      c #73FCBA",
"'      c #1ED31A",
")      c #202020",
"!      c #56FC45",
"~      c #04EB66",
"{      c #2DFC2D",
"]      c #55FC39",
"^      c #04EB1A",
"/      c #2BFC5C",
"(      c #1FFC19",
"_      c #58AF4A",
":      c #0F7702",
"<      c #223C1D",
"[      c #1AA308",
"}      c #1ACF02",
"|      c #0D7F02",
"1      c #021801",
"2      c #022001",
"3      c #215217",
"4      c #0A3302",
"                ",
"                ",
"                ",
"      ...       ",
"    .+@#@+.     ",
"    +$%&*%+     ",
"   .#=-;>,'.    ",
"   )!=~--&{.    ",
"   )]=^~~/(.    ",
"   ._!====:.    ",
"    <[]!}|1     ",
"     23342      ",
"                ",
"                ",
"                ",
"                "};

static char * red_led_xpm[] = {
"16 16 33 1",
"       c None",
".      c #020204",
"+      c #36020C",
"@      c #FE861C",
"#      c #541824",
"$      c #EE5A04",
"%      c #820214",
"&      c #B24A5C",
"*      c #A60820",
"=      c #B2021C",
"-      c #FE3A5C",
";      c #FEDEC4",
">      c #CA0224",
",      c #FE2E34",
"'      c #E63E04",
")      c #FE4764",
"!      c #202020",
"~      c #EE6F04",
"{      c #D20224",
"]      c #220204",
"^      c #D61A24",
"/      c #FE9E4C",
"(      c #3E1E24",
"_      c #7A0214",
":      c #FE1A2C",
"<      c #EE0224",
"[      c #1A0204",
"}      c #FEB274",
"|      c #EE1204",
"1      c #FE562C",
"2      c #4B020A",
"3      c #96021C",
"4      c #FE7604",
"                ",
"                ",
"                ",
"      ...       ",
"    .23=32.     ",
"    2>'/;'2     ",
"   .=<~4@}^.    ",
"   !)<$~~/,.    ",
"   !-<|$$1:.    ",
"   .&)<<<<_.    ",
"    (*-){%[     ",
"     ]##+]      ",
"                ",
"                ",
"                ",
"                "};

typedef struct host_info {

    char* name;
    char* ip;
    int port;
    struct host_info* next;

} host_info;

static host_info* host_list = NULL;

typedef struct gtk_gui {

    GtkWidget *window;
    GtkWidget *vbox;

    GtkWidget *hbox;
    GtkWidget *option;
    GtkWidget *command_menu;
    GtkWidget *command_item;
    GtkWidget *status_led;    
    GdkPixmap *online_led_pixmap;
    GdkBitmap *online_led_mask;
    GdkPixmap *offline_led_pixmap;
    GdkBitmap *offline_led_mask;

    GtkWidget *entry1;
    GtkWidget *entry2;
    GtkWidget *text;

    gint selected_command;
    char *cur_user;

} gtk_gui;
static gtk_gui gui;

void get_messages(int *sockfd);
void process_message(char *data);
int send_data(char *to_host,char *data);
void add_host(char *name,char *ip,int port);
host_info *get_host(char *name);
void remove_host(char *name);
void make_gui(int *argc,char **argv[]);
GtkWidget *make_menu_item(char *name,GtkSignalFunc callback,gpointer data);
void select_command(GtkWidget *widget,command_int cmnd);
gboolean delete_event(GtkWidget *widget,GdkEvent *event,gpointer data);
void destroy(GtkWidget *widget);
void enter_callback(GtkWidget *widget, GtkWidget *entry);

int main(int argc, char *argv[]) {

    int sockfd;
    struct sockaddr_in client_addr;
    pthread_t listener;

    if(argc != 5) {
	fprintf(stderr,"usage: %s client_ip client_port server_ip server_port\n",argv[0]);
	exit(1);
    }

    if((sockfd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
	perror("socket");
	exit(2);
    }
    client_addr.sin_family = AF_INET;
    client_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    client_addr.sin_port = htons(atoi(argv[2]));
    
    if(bind(sockfd,(struct sockaddr*)&client_addr,sizeof(client_addr)) < 0) {
	perror("bind");
	exit(3);
    }
    
    if(listen(sockfd, 5) < 0) {
	perror("listen");
	exit(4);
    }

    add_host("epos_server",argv[3],atoi(argv[4]));
    add_host("epos_client",argv[1],atoi(argv[2]));

    pthread_create(&listener,NULL,(void*)&get_messages,&sockfd);

    make_gui(&argc,&argv);

    while(host_list != NULL) {
	remove_host(host_list->name);
    }

    return 0;   
    
}

void get_messages(int *socket) {

    int sockfd = *socket;
    int client_sockfd,len;
    int data_len,data_cnt;
    struct sockaddr_in client_addr;
    char *received;

    while(1) {
	printf("waiting for connection\n");
	if((client_sockfd = 
	    accept(sockfd,(struct sockaddr*)&client_addr,&len)) < 0 ) {
	    perror("accept");
	    exit(6);
	}
	printf("new message\n");
	memset(buf,'\0',sizeof(char)*(BUFFER_SIZE));
	read(client_sockfd, buf,8);
	data_cnt = 0;
	data_len = atoi(buf) + 1;
	printf("len %s = %d\n",buf,data_len);
	
	received = (char*)malloc(sizeof(char)*(data_len+1));
	memset(received,'\0',sizeof(char)*(data_len+1));
	
	while(data_len - data_cnt) {
	    printf("%d - %d = %d\n",data_len,data_cnt,data_len-data_cnt);
	    memset(buf,'\0',sizeof(char)*BUFFER_SIZE);
	    data_cnt += read(client_sockfd,buf,BUFFER_SIZE);
	    strcat(received,buf);
	}
	
	process_message(received);
	
	close(client_sockfd);
	
	free(received);
    }

}

void process_message(char *data) {

    int n;
    char *command,*message[MAX_DATA_FIELDS];

    printf("got message: %s\n",data);

    command = strtok(data,SEPARATION_STRING);
    for(n=0;n<MAX_DATA_FIELDS;n++)
	message[n] = strtok(NULL,SEPARATION_STRING);

    if(strcmp(command,command_i2s(SRV_RPLY)) == 0) {
	if(strcmp(message[0],AFFIRM_REGISTER) == 0) {
	    printf("%s: %s\n",message[0],message[1]);
	} else if(strcmp(message[0],DENY_REGISTER) == 0) {
	    printf("%s: %s\n",message[0],message[1]);
	}
	if(strcmp(message[0],AFFIRM_LOGIN) == 0) {
	    printf("%s: %s\n",message[0],message[1]);
	    gtk_pixmap_set(GTK_PIXMAP(gui.status_led),
			   gui.online_led_pixmap,
			   gui.online_led_mask);
	    if(gui.cur_user != NULL) {
	        free(gui.cur_user);
	    }
	    gui.cur_user = (char*)malloc(sizeof(char)*
	        (strlen(gtk_entry_get_text(GTK_ENTRY(gui.entry1)))+1));
	    sprintf(gui.cur_user,"%s",
		    gtk_entry_get_text(GTK_ENTRY(gui.entry1)));
	} else if(strcmp(message[0],DENY_LOGIN) == 0) {
	    printf("%s: %s\n",message[0],message[1]);
	}
	if(strcmp(message[0],AFFIRM_LOGOUT) == 0) {
	    printf("%s: %s\n",message[0],message[1]);
	    gtk_pixmap_set(GTK_PIXMAP(gui.status_led),
			   gui.offline_led_pixmap,
			   gui.offline_led_mask);
	    if(gui.cur_user != NULL) {
	        free(gui.cur_user);
	    }
	    gui.cur_user = NULL;
	} else if(strcmp(message[0],DENY_LOGOUT) == 0) {
	    printf("%s: %s\n",message[0],message[1]);
	}
    } else {

	gtk_text_freeze(GTK_TEXT(gui.text));
	gtk_text_insert(GTK_TEXT(gui.text),NULL,NULL,NULL,message[0],-1);
	gtk_text_insert(GTK_TEXT(gui.text),NULL,NULL,NULL,": ",-1);
	gtk_text_insert(GTK_TEXT(gui.text),NULL,NULL,NULL,message[2],-1);
	gtk_text_insert(GTK_TEXT(gui.text),NULL,NULL,NULL,"\n",-1);
	gtk_text_thaw(GTK_TEXT(gui.text));

    }

}

void make_gui(int *argc,char **argv[]) {

    int n;
    GtkStyle *style;

    gtk_init(argc,argv);

    gui.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(gui.window),"Epos");
    gtk_signal_connect(GTK_OBJECT(gui.window),"delete_event",
                       GTK_SIGNAL_FUNC(delete_event),NULL);
    gtk_signal_connect(GTK_OBJECT(gui.window),"destroy",
                       GTK_SIGNAL_FUNC(destroy),NULL);

    gui.vbox = gtk_vbox_new(FALSE,0);
    gtk_container_add(GTK_CONTAINER(gui.window),gui.vbox);
    gtk_widget_show(gui.vbox);

    gui.text = gtk_text_new(NULL,NULL);
    gtk_text_set_editable(GTK_TEXT(gui.text),FALSE);
    gtk_box_pack_start(GTK_BOX(gui.vbox),gui.text,FALSE,FALSE,0);
    gtk_widget_show(gui.text);

    gui.hbox = gtk_hbox_new(FALSE,0);
    gtk_container_add(GTK_CONTAINER(gui.window),gui.hbox);
    gtk_widget_show(gui.hbox);
        
    gui.option = gtk_option_menu_new();
    gui.command_menu = gtk_menu_new();
    for(n=0;n<__N_COMMANDS;n++) {
	gui.command_item = make_menu_item(command_i2s(n),
					  GTK_SIGNAL_FUNC(select_command),
					  GINT_TO_POINTER(n));
	gtk_menu_append(GTK_MENU(gui.command_menu),gui.command_item);
    }
    gtk_menu_set_active(GTK_MENU(gui.command_menu),LOGIN);
    gui.selected_command = LOGIN;
    gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.option),gui.command_menu);
    gtk_box_pack_start(GTK_BOX(gui.hbox),gui.option,FALSE,FALSE,0);
    gtk_widget_show(gui.option);

    gtk_box_pack_start(GTK_BOX(gui.vbox),gui.hbox,FALSE,FALSE,0);

    gui.entry1 = gtk_entry_new();
    gtk_signal_connect(GTK_OBJECT(gui.entry1),"activate",
                       GTK_SIGNAL_FUNC(enter_callback),gui.entry1);
    gtk_box_pack_start(GTK_BOX(gui.vbox),gui.entry1,FALSE,FALSE,0);
    gtk_widget_show(gui.entry1);

    gui.entry2 = gtk_entry_new();
    gtk_signal_connect(GTK_OBJECT(gui.entry2),"activate",
                       GTK_SIGNAL_FUNC(enter_callback),gui.entry2);
    gtk_box_pack_start(GTK_BOX(gui.vbox),gui.entry2,FALSE,FALSE,0);
    gtk_widget_show(gui.entry2);    

    gtk_widget_show(gui.window);

    style = gtk_widget_get_style(gui.text);
    gui.online_led_pixmap = 
	gdk_pixmap_create_from_xpm_d(gui.window->window,&gui.online_led_mask,
				     &style->bg[GTK_STATE_NORMAL],
				     green_led_xpm);
    gui.offline_led_pixmap = 
	gdk_pixmap_create_from_xpm_d(gui.window->window,&gui.offline_led_mask,
				     &style->bg[GTK_STATE_NORMAL],
				     red_led_xpm);
    gui.status_led = gtk_pixmap_new(gui.offline_led_pixmap,
				    gui.offline_led_mask);
    gtk_box_pack_start(GTK_BOX(gui.hbox),gui.status_led,FALSE,FALSE,0);
    gtk_widget_show(gui.status_led);

    gtk_main();

}

GtkWidget *make_menu_item(char *name,GtkSignalFunc callback,gpointer data) {

    GtkWidget *item;

    item = gtk_menu_item_new_with_label(name);
    if(callback != NULL) {
	gtk_signal_connect(GTK_OBJECT(item),"activate",callback,data);
    }
    gtk_widget_show(item);

    return item;

}

void select_command(GtkWidget *widget,command_int cmnd) {
  
  char *message;
    gui.selected_command = cmnd;

    switch(cmnd) {
    case REGISTER:
      	gtk_widget_show(gui.entry1);
	gtk_widget_show(gui.entry2);
	break;
    case CHPASSWD:
	gtk_widget_show(gui.entry1);
	gtk_widget_show(gui.entry2);
	break;
    case LOGIN:
	gtk_widget_show(gui.entry1);
	gtk_widget_show(gui.entry2);
	break;
    case LOGOUT:
	gtk_widget_hide(gui.entry1);
	gtk_widget_hide(gui.entry2);
	if(gui.cur_user != NULL) {
	    message = (char*)malloc(sizeof(char)*(8+strlen(gui.cur_user)+2));
	    sprintf(message,"%8s%s%s",
		    command_i2s(LOGOUT),SEPARATION_STRING,gui.cur_user);
	    send_data("epos_server",message);
	    free(message);
	}
	break;
    case SRV_MSG:
	gtk_widget_show(gui.entry1);
	gtk_widget_show(gui.entry2);
	break;
    default:
	break;
    }

}

gboolean delete_event(GtkWidget *widget,GdkEvent *event,gpointer data) {

    gtk_widget_hide(gui.window);
    return FALSE; 

}

void destroy(GtkWidget *widget) {

    gtk_main_quit();

}

void enter_callback(GtkWidget *widget,GtkWidget *entry) {
    
    char *message,port_string[7];
    int len = 0,entry1 = 0,entry2 = 0;
    host_info *me = NULL;
    
    char *to_host = "epos_host";
    

    if(gui.cur_user != NULL) {
	len += strlen(gui.cur_user) + 1;
    } else {
	gui.cur_user = (char*)malloc(sizeof(char)*
	    (strlen(gtk_entry_get_text(GTK_ENTRY(gui.entry1)))+1));
	sprintf(gui.cur_user,"%s",gtk_entry_get_text(GTK_ENTRY(gui.entry1)));
	len += strlen(gui.cur_user) + 1;
    }
    len += strlen(to_host) + 1;

    if((entry1 = GTK_WIDGET_VISIBLE(gui.entry1))) {
	len += strlen(gtk_entry_get_text(GTK_ENTRY(gui.entry1))) + 1;
    }
    if((entry2 = GTK_WIDGET_VISIBLE(gui.entry2))) {
	len += strlen(gtk_entry_get_text(GTK_ENTRY(gui.entry2))) + 1;
    }
    
    switch(gui.selected_command) {
    case REGISTER:
	me = get_host("epos_client");
	sprintf(port_string,"%d",me->port);
	len += strlen(me->ip) + strlen(port_string) + 2;
	break;
    case CHPASSWD:
	len += strlen(gui.cur_user) + 1;
    case LOGIN:
	me = get_host("epos_client");
	sprintf(port_string,"%d",me->port);
	len += strlen(me->ip) + strlen(port_string) + 2;	
	break;
    case LOGOUT:
	break;
    case SRV_MSG:
	break;
    default:
	break;
    }

    message = (char*)malloc(sizeof(char)*(len+8+1));
    memset(message,'\0',sizeof(char)*(len+8+1));

    sprintf(message,"%8s",command_i2s(gui.selected_command));

    strcat(message,gui.cur_user);
    strcat(message,SEPARATION_STRING);

    strcat(message,to_host);
    strcat(message,SEPARATION_STRING);

    if(entry1) {
	strcat(message,gtk_entry_get_text(GTK_ENTRY(gui.entry1)));
	if(entry2) {
	    strcat(message,SEPARATION_STRING);
	    strcat(message,gtk_entry_get_text(GTK_ENTRY(gui.entry2)));
	}
	if(me != NULL) {
	    strcat(message,SEPARATION_STRING);
	    strcat(message,me->ip);
	    strcat(message,SEPARATION_STRING);
	    strcat(message,port_string);
	}
    }
    printf("about to send: [%s]\n",message);

    send_data("epos_server",message);

    free(message);
    
}

int send_data(char *to_host, char *data) {

    int sockfd,nleft,nsent,nchars,n;
    struct sockaddr_in serv_addr;
    struct hostent *hp;
    host_info *host;

    if((host = get_host(to_host)) == NULL) {
	fprintf(stderr,"%s does not exist\n",to_host);
    }

    if((sockfd = socket(AF_INET,SOCK_STREAM, 0)) < 0) {
	perror("socket");
	exit(2);
    }
   
    serv_addr.sin_family = AF_INET;
    hp = gethostbyname(host->ip);
    bcopy(hp->h_addr,(char*)&serv_addr.sin_addr,hp->h_length);
    serv_addr.sin_port = htons(host->port);

    if(connect(sockfd,(struct sockaddr*)&serv_addr,sizeof(serv_addr)) < 0) {
	perror("connect");
	exit(3);
    }

    sprintf(buf,"%8d",strlen(data));
    write(sockfd,buf,8);

    nsent = 0;
    nleft = strlen(data)+1;
    while(nleft) {
	nchars = (nleft < BUFFER_SIZE) ? nleft : BUFFER_SIZE;
	for(n=0;n<nchars;n++) buf[n] = data[n+nsent];
	write(sockfd,buf,nchars);
	nleft -= nchars;
	nsent += nchars;
    }

    close(sockfd);

    return nsent;

}

void add_host(char* name,char* ip,int port) {

    host_info* walker = host_list;
    host_info* new;

    if(get_host(name) == NULL) {
	new = (host_info*)malloc(sizeof(host_info));
	new->name = (char*)malloc(sizeof(char)*(strlen(name)+1));
	strcpy(new->name,name);
	new->ip = (char*)malloc(sizeof(char)*(strlen(ip)+1));
	strcpy(new->ip,ip);
	new->port = port;
	new->next = NULL;
	
	if(walker == NULL) {
	    host_list = new;
	} else {
	    while(walker->next != NULL) {
		walker = walker->next;
	    }
	    walker->next = new;
	}
    }
    return;
}

host_info *get_host(char *name) {

    host_info *walker = host_list;

    while(walker != NULL) {
	if(strcmp(name,walker->name) == 0) {
	    return walker;
	}
	walker = walker->next;
    }
    return NULL;

}

void remove_host(char* name) {

    host_info *walker = host_list, *tmp;

    if(walker == NULL) {
	return;
    } else if(strcmp(name,walker->name) == 0) {
	free(walker->name);
	free(walker->ip);
	host_list = walker->next;
	free(walker);
	walker = NULL;
	return;
    } else {
	while(walker->next != NULL) {
	    if(strcmp(walker->next->name,name) == 0) {
		free(walker->next->name);
		free(walker->next->ip);
		tmp = walker->next;
		walker->next = walker->next->next;
		free(tmp);
		tmp = NULL;
		return;
	    }
	    walker = walker->next;
	}
    }
    return;

}
