#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <strings.h>
#include "epos_protocol.h"

static char buf[BUFFER_SIZE];

typedef struct client_record {

    char* name;
    char* password;
    char* ip;
    int port;
    struct client_record* next;

} client_record;

static client_record* client_list = NULL;

void serve(int sockfd);
void process_command(char *data);
int send_data(char *to_client,char* data);
int send_data_to(char *ip,int port,char* data);
int check_client_password(char* name,char* password);
int set_client_addr(char* name,char* ip,int port);
int add_client_record(char* name,char* password,char* ip,int port);
client_record *get_client_record(char *name);
int remove_client_record(char* name,char* password);
void print_client_records(void);

int main(int argc, char* argv[]) {

    int sockfd;
    struct sockaddr_in serv_addr;

    if(argc != 2) {
	fprintf(stderr,"usage: %s port\n",argv[0]);
	exit(1);
    }

    if((sockfd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
	perror("socket");
	exit(2);
    }
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_addr.sin_port = htons(atoi(argv[1]));
    
    if(bind(sockfd,(struct sockaddr*)&serv_addr,sizeof(serv_addr)) < 0) {
	perror("bind");
	exit(3);
    }
    
    if(listen(sockfd, 5) < 0) {
	perror("listen");
	exit(4);
    }

    serve(sockfd);
    
    return 0;
}

void serve(int sockfd) {

    int client_sockfd,len;
    int data_len,data_cnt;
    struct sockaddr_in client_addr;
    char *received;
    
    while(1) {
	printf("waiting for connection\n");
	if((client_sockfd = 
	    accept(sockfd,(struct sockaddr*)&client_addr,&len)) < 0 ) {
	    perror("accept");
	    exit(6);
	}
	printf("new client\n");
	memset(buf,'\0',sizeof(char)*(BUFFER_SIZE));
	read(client_sockfd, buf,8);
	data_cnt = 0;
	data_len = atoi(buf) + 1;
	printf("len %s = %d\n",buf,data_len);

	received = (char*)malloc(sizeof(char)*(data_len+1));
	memset(received,'\0',sizeof(char)*(data_len+1));
	
	while(data_len - data_cnt) {
	    printf("%d - %d = %d\n",data_len,data_cnt,data_len-data_cnt);
	    memset(buf,'\0',sizeof(char)*BUFFER_SIZE);
	    data_cnt += read(client_sockfd,buf,BUFFER_SIZE);
	    strcat(received,buf);
	}

	printf("received: %s\n",received);

	process_command(received);

	close(client_sockfd);

	free(received);
    }

}

void process_command(char *data) {

    char cmnd[9],*fields[MAX_DATA_FIELDS],*message;
    int n,command_nr,null_field = 0;
    client_record *client_rec;

    printf("data: %s\n",data);

    for(n=0;n<8;n++) {
	cmnd[n] = data[n];
	printf("%c %c %d %d\n",cmnd[n],data[n],cmnd[n],data[n]);
    }
    cmnd[8] = '\0';
    command_nr = command_s2i(cmnd);
    data += 8;

    n = 0;
    fields[n] = strtok(data,SEPARATION_STRING);
    printf("%d: %s\n",n,fields[n]);
    while(fields[n] != NULL) {
	fields[++n] = strtok(NULL,SEPARATION_STRING);
	if(fields[n])
	  printf("%d: %s\n",n,fields[n]);
    }

    switch(command_nr) {
    case REGISTER:
	for(n=0;n<6;n++) {
	    if(fields[n] == NULL) {
		null_field = 1;
	    }
	}
	if(!null_field) {
	    printf("register: %s %s %s %s\n",
		   fields[0],fields[3],fields[4],fields[5]);
	    if(add_client_record(fields[0],fields[3],
				 fields[4],atoi(fields[5]))) {
		message = (char*)malloc(sizeof(char)*
					(strlen(command_i2s(SRV_RPLY))+
					 strlen(AFFIRM_REGISTER)+
					 strlen(fields[0])+
					 3));
		sprintf(message,"%8s%s%s%s%s",
			command_i2s(SRV_RPLY),SEPARATION_STRING,
			AFFIRM_REGISTER,SEPARATION_STRING,
			fields[0]);
		send_data(fields[0],message);
		free(message);
	    } else {
		message = (char*)malloc(sizeof(char)*
					(strlen(command_i2s(SRV_RPLY))+
					 strlen(DENY_REGISTER)+
					 strlen(fields[0])+
					 3));
		sprintf(message,"%8s%s%s%s%s",
			command_i2s(SRV_RPLY),SEPARATION_STRING,
			DENY_REGISTER,SEPARATION_STRING,
			fields[0]);
		send_data_to(fields[4],atoi(fields[5]),message);
		free(message);
	    }
		print_client_records();
	}
	break;
    case LOGIN:
	for(n=0;n<6;n++) {
	    if(fields[n] == NULL) {
		null_field = 1;
	    }
	}
	if(!null_field) {
	    if(check_client_password(fields[0],fields[3]) &&
	       set_client_addr(fields[0],fields[4],atoi(fields[5]))) {
	        printf("ok...\n");
		message = (char*)malloc(sizeof(char)*
					(strlen(command_i2s(SRV_RPLY))+
					 strlen(AFFIRM_LOGIN)+
					 strlen(fields[0])+
					 3));
		sprintf(message,"%8s%s%s%s%s",
			command_i2s(SRV_RPLY),SEPARATION_STRING,
			AFFIRM_LOGIN,SEPARATION_STRING,
			fields[0]);
		send_data_to(fields[4],atoi(fields[5]),message);
		free(message);
	    } else {
	      printf("not ok...\n");
		message = (char*)malloc(sizeof(char)*
					(strlen(command_i2s(SRV_RPLY))+
					 strlen(DENY_LOGIN)+
					 strlen(fields[0])+
					 3));
		sprintf(message,"%8s%s%s%s%s",
			command_i2s(SRV_RPLY),SEPARATION_STRING,
			DENY_LOGIN,SEPARATION_STRING,
			fields[0]);
		send_data_to(fields[4],atoi(fields[5]),message);
		free(message);
	    }
	    print_client_records();
	}
	break;
    case LOGOUT:
	for(n=0;n<1;n++) {
	    if(fields[n] == NULL) {
		null_field = 1;
	    }
	}
	if(!null_field) {
	    if(get_client_record(fields[0]) != NULL) {
		printf("ok...\n");
		message = (char*)malloc(sizeof(char)*
					(strlen(command_i2s(SRV_RPLY))+
					 strlen(AFFIRM_LOGOUT)+
					 strlen(fields[0])+
					 3));
		sprintf(message,"%8s%s%s%s%s",
			command_i2s(SRV_RPLY),SEPARATION_STRING,
			AFFIRM_LOGOUT,SEPARATION_STRING,
			fields[0]);
		send_data(fields[0],message);
		free(message);
		set_client_addr(fields[0],"",-1);
	    } else {
		printf("not ok...\n");
		message = (char*)malloc(sizeof(char)*
					(strlen(command_i2s(SRV_RPLY))+
					 strlen(DENY_LOGOUT)+
					 strlen(fields[0])+
					 3));
		sprintf(message,"%8s%s%s%s%s",
			command_i2s(SRV_RPLY),SEPARATION_STRING,
			DENY_LOGOUT,SEPARATION_STRING,
			fields[0]);
		send_data(fields[0],message);
		free(message);
	    }
	    print_client_records();
	}
	break;
    case SRV_MSG:
	for(n=0;n<4;n++) {
	    if(fields[n] == NULL) {
		null_field = 1;
	    }
	}
	if(!null_field) {
	    if(((client_rec = get_client_record(fields[2])) != NULL) &&
	       (strlen(client_rec->ip) > 0)) {
	        message = (char*)malloc(sizeof(char)*
					(strlen(command_i2s(SRV_MSG))+
					 strlen(fields[0])+
					 strlen(fields[2])+
					 strlen(fields[3])+
					 4));
		sprintf(message,"%8s%s%s%s%s%s%s",
			command_i2s(SRV_MSG),SEPARATION_STRING,
			fields[0],SEPARATION_STRING,
			fields[2],SEPARATION_STRING,
			fields[3]);
		send_data(fields[2],message);
		free(message);
	    }
	} else {
	    message = (char*)malloc(sizeof(char)*
				    (strlen(command_i2s(SRV_RPLY))+
				     strlen(SRV_MSG_ERROR)+
				     strlen(fields[0])+
				     3));
	    sprintf(message,"%8s%s%s%s%s",
		    command_i2s(SRV_RPLY),SEPARATION_STRING,
		    SRV_MSG_ERROR,SEPARATION_STRING,
		    fields[0]);
	    send_data(fields[0],message);
	    free(message);
	}
	break;
    default:
	break;
    }

}

int send_data(char *to_client,char *data) {

    int rv;
    client_record *client_rec;

    if((client_rec = get_client_record(to_client)) == NULL) {
	fprintf(stderr,"%s does not exist\n",to_client);
	rv = 0;
    } else {
        rv = send_data_to(client_rec->ip,client_rec->port,data);
        printf("sent: [%s] to [%s]\n",data,to_client);
    }
    return rv;

}

int send_data_to(char *ip,int port, char *data) {

    int sockfd,nleft,nsent,nchars,n;
    struct sockaddr_in serv_addr;
    struct hostent *hp;

    if((sockfd = socket(AF_INET,SOCK_STREAM, 0)) < 0) {
	perror("socket");
	exit(2);
    }
   
    serv_addr.sin_family = AF_INET;
    hp = gethostbyname(ip);
    bcopy(hp->h_addr,(char*)&serv_addr.sin_addr,hp->h_length);
    serv_addr.sin_port = htons(port);

    if(connect(sockfd,(struct sockaddr*)&serv_addr,sizeof(serv_addr)) < 0) {
	perror("connect");
	exit(3);
    }

    sprintf(buf,"%8d",strlen(data));
    write(sockfd,buf,8);

    nsent = 0;
    nleft = strlen(data)+1;
    while(nleft) {
	nchars = (nleft < BUFFER_SIZE) ? nleft : BUFFER_SIZE;
	for(n=0;n<nchars;n++) buf[n] = data[n+nsent];
	write(sockfd,buf,nchars);
	nleft -= nchars;
	nsent += nchars;
	printf("\n");
    }
    close(sockfd);

    return nsent;

}

int set_client_addr(char* name,char* ip,int port) {

    client_record *client_rec;

    if((client_rec = get_client_record(name)) == 0) {
	return 0;
    } else {
	if(client_rec->ip != NULL) {
	    free(client_rec->ip);
	}
	client_rec->ip = (char*)malloc(sizeof(char)*(strlen(ip)+1));
	strcpy(client_rec->ip,ip);
	client_rec->port = port;
	return 1;
    }

}

int check_client_password(char* name,char* password) {
    
    client_record* who;

    if((who = get_client_record(name)) == NULL)	return 0;

    return !(strcmp(who->password,password));    

}

int add_client_record(char* name,char* password,char* ip,int port) {
   
    client_record* walker = client_list;
    client_record* new;
    int rv = 0;
    
    if(get_client_record(name) == NULL) {
	new = (client_record*)malloc(sizeof(client_record));
	new->name = (char*)malloc(sizeof(char)*(strlen(name)+1));
	strcpy(new->name,name);
	new->password = (char*)malloc(sizeof(char)*(strlen(password)+1));
	strcpy(new->password,password);
	new->ip = (char*)malloc(sizeof(char)*(strlen(ip)+1));
	strcpy(new->ip,ip);
	new->port = port;
	new->next = NULL;
	
	rv = 1;

	if(walker == NULL) {
	    client_list = new;
	} else {
	    while(walker->next != NULL) {
		walker = walker->next;
	    }
	    walker->next = new;
	}
    }

    return rv;
}

client_record *get_client_record(char *name) {

    client_record *walker = client_list;

    while(walker != NULL) {
	if(strcmp(name,walker->name) == 0) {
	    return walker;
	}
	walker = walker->next;
    }
    return NULL;

}

int remove_client_record(char* name,char* password) {

    client_record *walker = client_list, *tmp;

    if(walker == NULL) {
	return 0;
    } else if(strcmp(name,walker->name) == 0) {
	if(strcmp(password,walker->password) == 0) {
	    free(walker->name);
	    free(walker->ip);
	    client_list = walker->next;
	    free(walker);
	    walker = NULL;
	    return 1;
	}
    } else {
	while(walker->next != NULL) {
	    if(strcmp(walker->next->name,name) == 0) {
		if(strcmp(password,walker->next->password) == 0) {
		    free(walker->next->name);
		    free(walker->next->ip);
		    tmp = walker->next;
		    walker->next = walker->next->next;
		    free(tmp);
		    tmp = NULL;
		    return 1;
		}
	    }
	    walker = walker->next;
	}
    }
    return 0;

}

void print_client_records(void) {

    client_record *client_rec = client_list;

    printf("client list:\n");
    while(client_rec != NULL) {
	printf("%s %s %s %d\n",
	       client_rec->name,client_rec->password,
	       client_rec->ip,client_rec->port);
	client_rec = client_rec->next;
    }

}
